/*
 * Papillon - Solaris security module -  http://www.roqe.org/papillon
 * $Id: chroot.c,v 1.14 2002/09/02 09:33:41 kr Exp $
 *
 * Chroot protection. Initial implementation by Heiko Krupp. Ideas
 * have been adopted from the HAP-Linux kernel patch, chroot(), mount()
 * and mknod() interception.
 */

#define _KERNEL

#include "headers.h"

#if defined(CHROOTPROT)

extern pap_syscall_t syscalls[];
#ifdef _64BIT
extern pap_syscall_t syscalls32[];
#endif

extern pap_config_t config;
extern krwlock_t config_rwlock;

int64_t pap_chroot(const char *path)
{
    int64_t ret;

    inc_refcnt();
    dcmn_err((CE_CONT, "chroot() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    /*
     * Skip if no chroot protection or no new root set for process
     */
    if (!config.chrootprot || !u.u_rdir)
	goto skip;

    if (config.chrootprot == DENY) {
	cmn_err(CE_WARN, "Denied calling duplicate chroot() from %s "
		"(pid %d, uid %d).",
		u.u_psargs, ddi_get_pid(), CRED()->cr_ruid);
	ret = set_errno(EPERM);
    } else {
	cmn_err(CE_WARN, "Calling duplicate chroot() from %s "
		"(pid %d, uid %d).",
		u.u_psargs, ddi_get_pid(), CRED()->cr_ruid);
      skip:
	ret = syscalls[CHROOT].sc(path);
    }

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}

int64_t pap_mount(const char *spec, const char *dir, int mflag, char
		  *fstype, char *dataptr, int datalen, char *optptr, int
		  optlen)
{
    char k_path[256] = "<none>";
    int64_t ret;

    inc_refcnt();
    dcmn_err((CE_CONT, "mount() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    /*
     * Skip if no chroot protection or no new root set for process
     */
    if (!config.chrootprot || !u.u_rdir)
	goto skip;

    copyinstr(dir, k_path, 255, (size_t *) & ret);

    if (config.chrootprot == DENY) {
	cmn_err(CE_WARN,
		"Denied mounting %s in chroot env (pid %d, uid %d).", k_path,
		ddi_get_pid(), CRED()->cr_ruid);
	ret = set_errno(EPERM);
    } else {
	cmn_err(CE_WARN, "Mounting %s in chroot env (pid %d, uid %d).",
		k_path, ddi_get_pid(), CRED()->cr_ruid);
      skip:
	ret = syscalls[MOUNT].sc(spec, dir, mflag, fstype, dataptr, datalen,
				 optptr, optlen);
    }

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}

int64_t pap_mknod(const char *path, mode_t mode, dev_t dev)
{
    char k_path[256] = "<none>";
    int64_t ret;

    inc_refcnt();
    dcmn_err((CE_CONT, "mknod() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    /*
     * Skip if no chroot protection or no new root set for process and
     * creation of a charakter/block device requested.
     */
    if (!config.chrootprot || !(u.u_rdir && (mode | S_IFCHR | S_IFBLK)))
	goto skip;

    copyinstr(path, k_path, 255, (size_t *) & ret);

    if (config.chrootprot == DENY) {
	cmn_err(CE_WARN, "Denied making device node c|b %s in chroot env "
		"(pid %d, uid %d).", k_path, ddi_get_pid(), CRED()->cr_ruid);
	ret = set_errno(EPERM);
    } else {
	cmn_err(CE_WARN, "Making device node c|b %s in chroot env "
		"(pid %d, uid %d).", k_path, ddi_get_pid(), CRED()->cr_ruid);
      skip:
	ret = syscalls[MKNOD].sc(path, mode, dev);
    }

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}

int64_t pap_modctl(int cmd, uintptr_t a1, uintptr_t a2, uintptr_t a3,
		   uintptr_t a4, uintptr_t a5)
{
    int64_t ret;

    /*
     * If not modloading jump directly to the original code. Don't
     * reference count or lock this syscall! If you do, you can't unload
     * Papillon, think twice!
     */
    if (cmd != MODLOAD)
	return syscalls[MODCTL].sc(cmd, a1, a2, a3, a4, a5);

    inc_refcnt();
    dcmn_err((CE_CONT, "() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    /*
     * Skip if no chroot protection or no new root set for process and
     * and loading a new module.
     */
    if (!config.chrootprot || !u.u_rdir)
	goto skip;

    if (config.chrootprot == DENY) {
	cmn_err(CE_WARN, "Denied loading kernel module in chroot env "
		"(pid %d, uid %d).", ddi_get_pid(), CRED()->cr_ruid);
	ret = set_errno(EPERM);
    } else {
	cmn_err(CE_WARN, "Loading kernel module in chroot env "
		"(pid %d, uid %d).", ddi_get_pid(), CRED()->cr_ruid);
      skip:
	ret = syscalls[MODCTL].sc(cmd, a1, a2, a3, a4, a5);
    }

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}

int64_t pap_chmod(const char *path, mode_t mode)
{
    char k_path[256] = "<none>";
    int64_t ret;

    inc_refcnt();
    dcmn_err((CE_CONT, "chmod() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    /*
     * Skip if no chroot protection or no new root set for process and
     * mode +s.
     */
    if (!config.chrootprot || !(u.u_rdir && (mode | S_ISUID | S_ISGID)))
	goto skip;

    copyinstr(path, k_path, 255, (size_t *) & ret);

    if (config.chrootprot == DENY) {
	cmn_err(CE_WARN, "Denied changing permission +s %s in chroot env "
		"(pid %d, uid %d).", k_path, ddi_get_pid(), CRED()->cr_ruid);
	ret = set_errno(EPERM);
    } else {
	cmn_err(CE_WARN, "Changing permssion +s %s in chroot env "
		"(pid %d, uid %d).", k_path, ddi_get_pid(), CRED()->cr_ruid);
      skip:
	ret = syscalls[CHMOD].sc(path, mode);
    }

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}

int64_t pap_xmknod(const char *path, mode_t mode, dev_t dev)
{
    char k_path[256] = "<none>";
    int64_t ret;

    inc_refcnt();
    dcmn_err((CE_CONT, "xmknod() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    /*
     * Skip if no chroot protection or no new root set for process and
     * creation of a charakter/block device requested.
     */
    if (!config.chrootprot || !(u.u_rdir && (mode | S_IFCHR | S_IFBLK)))
	goto skip;

    copyinstr(path, k_path, 255, (size_t *) & ret);

    if (config.chrootprot == DENY) {
	cmn_err(CE_WARN, "Denied making device node c|b %s in chroot env "
		"(pid %d, uid %d).", k_path, ddi_get_pid(), CRED()->cr_ruid);
	ret = set_errno(EPERM);
    } else {
	cmn_err(CE_WARN, "Making device node c|b %s in chroot env "
		"(pid %d, uid %d).", k_path, ddi_get_pid(), CRED()->cr_ruid);
      skip:
	ret = syscalls[XMKNOD].sc(path, mode, dev);
    }

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}



#ifdef _64BIT
int64_t pap_xmknod32(const char *path, mode_t mode, dev_t dev)
{
    char k_path[256] = "<none>";
    int64_t ret;

    inc_refcnt();
    dcmn_err((CE_CONT, "xmknod32() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    /*
     * Skip if no chroot protection or no new root set for process and
     * creation of a charakter/block device requested.
     */
    if (!config.chrootprot || !(u.u_rdir && (mode | S_IFCHR | S_IFBLK)))
	goto skip;

    copyinstr(path, k_path, 255, (size_t *) & ret);

    if (config.chrootprot == DENY) {
	cmn_err(CE_WARN, "Denied making device node c|b %s in chroot env "
		"(pid %d, uid %d).", k_path, ddi_get_pid(), CRED()->cr_ruid);
	ret = set_errno(EPERM);
    } else {
	cmn_err(CE_WARN, "Making device node c|b %s in chroot env "
		"(pid %d, uid %d).", k_path, ddi_get_pid(), CRED()->cr_ruid);
      skip:
	ret = syscalls32[XMKNOD].sc(path, mode, dev);
    }

    rw_exit(&config_rwlock);
    dec_refcnt();

    return ret;
}

#endif				/* _64BIT */
#endif				/* CHROOTPROT */
