/*
 * Papillon - Solaris security module -  http://www.roqe.org/papillon
 * Copyright (c) 2000-2002 Konrad Rieck <kr@roqe.org>
 * $Id: papctl.c,v 1.14 2002/09/02 09:33:41 kr Exp $
 *
 * This is the Papillon control tool. It uses the unused syscall defined
 * in papillon.h in order to communicate with the module (even if it is
 * hidden). Most features can be toggles by using this tool when the 
 * module is loaded. Recompilation is not needed. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/syscall.h>

#include "headers.h"

pap_config_t c;

/*
 * Print the current status of the loaded module.
 */
void print_config()
{
    printf("\nCurrent configuration of the Papillon v%s module:\n"
	   "\n  - Features\n", VERSION);
#ifdef RSTPROC
    printf("    Restricted Proc:               %s\n", F(c.rstproc));
#endif
#ifdef PPROMISC
    printf("    Pseudo Promiscuous Flag:       %s\n", F(c.ppromisc));
#endif
#ifdef MODHIDING
    printf("    Module Hiding:                 %s\n", F(c.modhiding));
#endif
#ifdef SECSTDFD
    printf("    Secure STDIO File Descriptors: %s\n", F(c.secstdfd));
#endif

    printf("\n  - Protections\n");
#ifdef SYMPROT
    printf("    Symlink Protection:            %s\n", P(c.symprot));
#endif
#ifdef HARDPROT
    printf("    Hardlink Protection:           %s\n", P(c.hardprot));
#endif
#ifdef FIFOPROT
    printf("    Fifo Protection:               %s\n", P(c.fifoprot));
#endif
#ifdef CHROOTPROT
    printf("    Chroot Protection:             %s\n", P(c.chrootprot));
#endif
    printf("\n");
}

/*
 * Print an error message according to the submited err_msg value.
 */
void print_error(int err_msg)
{
    printf("Error %d \n", err_msg);

    switch (err_msg) {
    case NO_AUTH:
	printf("You are not authorized to use this program.\n");
	break;
    case NOT_LOADED:
	printf("Papillon is not loaded. See modload(9E)\n");
	break;
    case INVALID_VALUE:
	printf("The assigned value is invalid for this type of variable\n");
	break;
    case INVALID_VARIABLE:
	printf("There is no such variable.\n");
	break;
    case MEM_FAULT:
	printf("Ex/Importing from user- to kernelspace failed.\n");
	break;
    case CONFIG_BLOCKED:
	printf("Configuration blocked. Volmgt running?\n");
	break;
    default:
	printf("Unknown error. *sigh*\n");
    }
    exit(EXIT_FAILURE);
}

void print_version(void)
{
    printf("Papillon v%s Control   "
	   "Copyright (c) 2000-2002 Konrad Rieck <kr@roqe.org>\n", VERSION);
}

/*
 * Guess what this does. 
 */
void print_usage()
{
    print_version();
    printf("Usage: papctl -s variable=value [variable=value ...]\n"
	   "       papctl -g | -v | -h\n"
	   "Options:\n"
	   "  -g                      get current configuration of the loaded module.\n"
	   "  -s variable=value ...   set current configuration of the loaded module.\n"
	   "  -h                      print this help.\n"
	   "  -v                      print version information.\n"
	   "\n"
	   " | Variable | Description                   | Values           |\n"
	   " |__________|_______________________________|__________________|\n"
#ifdef RSTPROC
	   " |    r     | Restricted Proc               |     on, off      |\n"
#endif
#ifdef PPROMISC
	   " |    p     | Pseudo Promiscuous Flag       |     on, off      |\n"
#endif
#ifdef MODHIDING
	   " |    m     | Module Hiding                 |     on, off      |\n"
#endif
#ifdef SECSTDFD
	   " |    i     | Secure STDIO File Descriptors |     on, off      |\n"
#endif
#ifdef SYMPROT
	   " |    s     | Symbolic Link Protection      | none, warn, deny |\n"
#endif
#ifdef HARDPROT
	   " |    h     | Hard Link Protection          | none, warn, deny |\n"
#endif
#ifdef FIFOPROT
	   " |    f     | FIFO Protection               | none, warn, deny |\n"
#endif
#ifdef CHROOTPROT
	   " |    c     | Chroot Protection             | none, warn, deny |\n"
#endif
	   "\n"
	   "Example (Features and protections may not be compiled in):\n"
	   "  # papctl -s f=warn r=off m=on s=deny\n");
    exit(EXIT_FAILURE);
}

/*
 * Set a protection according to the given string. 
 */
int protection(char *variable)
{
    char *value;
    value = variable + 2;

    if (!strcmp(value, "none"))
	return NONE;
    else if (!strcmp(value, "warn"))
	return WARN;
    else if (!strcmp(value, "deny"))
	return DENY;
    else {
	printf("Variable: %s\n", variable);
	print_error(INVALID_VALUE);
	return -1;
    }
}

/*
 * Set a feature according to the given string. 
 */
int feature(char *variable)
{
    char *value;
    value = variable + 2;

    if (!strcmp(value, "on"))
	return ON;
    else if (!strcmp(value, "off"))
	return OFF;
    else {
	printf("Variable: %s\n", variable);
	print_error(INVALID_VALUE);
	return -1;
    }
}

/*
 * Parse the strings gathered from the commandline
 */
void parse_variable(char *variable)
{
    if (variable[1] != '=')
	print_usage();

    switch ((int) variable[0]) {
#ifdef RSTPROC
    case 'r':
	c.rstproc = feature(variable);
	break;
#endif
#ifdef PPROMISC
    case 'p':
	c.ppromisc = feature(variable);
	break;
#endif
#ifdef MODHIDING
    case 'm':
	c.modhiding = feature(variable);
	break;
#endif
#ifdef SECSTDFD
    case 'i':
	c.secstdfd = feature(variable);
	break;
#endif
#ifdef SYMPROT
    case 's':
	c.symprot = protection(variable);
	break;
#endif
#ifdef HARDPROT
    case 'h':
	c.hardprot = protection(variable);
	break;
#endif
#ifdef FIFOPROT
    case 'f':
	c.fifoprot = protection(variable);
	break;
#endif
#ifdef CHROOTPROT
    case 'c':
	c.chrootprot = protection(variable);
	break;
#endif
    default:
	printf("Variable: %s\n", variable);
	print_error(INVALID_VARIABLE);
    }
}


int main(int argc, char *argv[])
{
    int request = 0, i, count = 0, ret;
    char **variables = NULL;

    if (getuid())
	print_error(NO_AUTH);

    if (argc < 2)
	print_usage();
    else if (!strcmp(argv[1], "-g"))
	request = GET;
    else if (!strcmp(argv[1], "-s"))
	request = SET;
    else if (!strcmp(argv[1], "-v")) {
	print_version();
	return (EXIT_SUCCESS);
    } else
	print_usage();

    signal(SIGSYS, print_error);

    switch (request) {
    case GET:

	if (argc > 2)
	    print_usage();
	if (syscall(SYS_papcomm, &c, GET))
	    print_error(MEM_FAULT);

	print_config();

	break;
    case SET:
	count = argc - 2;

	if (count < 1)
	    print_usage();

	if (syscall(SYS_papcomm, &c, GET))
	    print_error(MEM_FAULT);

	variables = (char **) malloc(sizeof(char *) * count);
	for (i = 0; i < count; i++) {
	    variables[i] = strdup(argv[i + 2]);
	    parse_variable(variables[i]);
	}

	ret = syscall(SYS_papcomm, &c, SET);

	if (ret)
	    print_error(ret);

	print_config();
	break;
    }

    return EXIT_SUCCESS;
}
